#include <cstdio>
using namespace std;

#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
using namespace cv;

const int FPS = 30;

class LoopedVideo : public VideoCapture {
public:
	LoopedVideo() {}
	virtual bool grab() {
		if(VideoCapture::grab()==false) {
			rewind();
		}
		return true;
	}
	void rewind() {
		set(CV_CAP_PROP_POS_FRAMES,0);
	}
};

VideoCapture camera;
Mat inputImage;
Mat currentBackgroundImage;
Mat savedBackgroundImage;
Mat differenceImage;

Mat grayscaleDifferenceImage;
Mat thresholdedDifferenceImage;
Mat fixedDifferenceImage;

Mat foregroundMask;
Mat foregroundImage;
int foregroundThreshold;

Mat backgroundImage;
const string staticBackgroundImagePath = "forest.jpg";
LoopedVideo backgroundMovie;
const string dynamicBackgroundMoviePath = "fishie.avi";

const string recordedMoviePath = "recorded.avi";
VideoWriter recordedMovie;
bool isRecording;

Mat resultImage;

void showUsage() {
	printf("Press 'b' to set background\n");
	printf("Press 'k' to terminate\n");
	printf("Press 'p' to save image\n");
	printf("Press 'r' to start and stop recording\n");
	printf("Press 'a' to increase foreground_threshold\n");
	printf("Press 'z' to decrease foreground_threshold\n");
}

void processKey(char key) {
	switch(key) {
		case 'b': 
			inputImage.copyTo(savedBackgroundImage);
			imshow("Background",savedBackgroundImage);
			break;
		case 'p':
			imwrite("image.jpg",inputImage);
			imwrite("diff.jpg",differenceImage);
			imwrite("foreraw.jpg",thresholdedDifferenceImage);
			imwrite("forefix.jpg",fixedDifferenceImage);
			imwrite("result.jpg",resultImage);
			break;
		case 'a':
			foregroundThreshold+=1;
			printf("foregroundThreshold = %d\n",foregroundThreshold);
			break;
		case 'z':
			if(foregroundThreshold>1) foregroundThreshold-=1;
			printf("foregroundThreshold = %d\n",foregroundThreshold);
			break;
		case 'r':
			if(isRecording == true) {
				isRecording = false;
				printf("Recording has been stoped!\n");
			}
			else {
				recordedMovie.open(recordedMoviePath,0,FPS,Size(640,480));
				isRecording = true;
				printf("Recording has started!\n");
			}
			break;
	}
}

int main(int argc, char *argv[]) {
	camera.open(0);
	if(!camera.isOpened()){
		printf("Cannot connect to the camera!\n"); 
		getchar();
	}
	else {
		printf("Connection established!\n");
		showUsage();
		backgroundMovie.open(dynamicBackgroundMoviePath);
		camera.set(CV_CAP_PROP_FRAME_WIDTH,640);
		camera.set(CV_CAP_PROP_FRAME_HEIGHT,480);
		char key=0;

		while(key!='k') {
			camera >> inputImage;
			backgroundMovie >> currentBackgroundImage;
			imshow("Camera Image",inputImage);
			key = waitKey(1000/FPS);
			processKey(key);
			if(savedBackgroundImage.data!=NULL) {
				absdiff(inputImage,savedBackgroundImage,differenceImage);
				imshow("Difference Image",differenceImage);
				
				cvtColor(differenceImage,grayscaleDifferenceImage,CV_BGR2GRAY);
				threshold(grayscaleDifferenceImage,thresholdedDifferenceImage,foregroundThreshold,255,THRESH_BINARY);
				imshow("Foreground Mask", thresholdedDifferenceImage);
				
				morphologyEx(thresholdedDifferenceImage,fixedDifferenceImage,MORPH_OPEN,Mat(),Point(-1,-1),1);
				morphologyEx(fixedDifferenceImage,fixedDifferenceImage,MORPH_CLOSE,Mat(),Point(-1,-1),3);
				imshow("Fixed Foreground Mask", fixedDifferenceImage);
				
				cvtColor(fixedDifferenceImage,foregroundMask,CV_GRAY2BGR);
				foregroundImage = foregroundMask & inputImage;
				//imshow("Foreground Image",foregroundImage);

				if(currentBackgroundImage.data!=NULL) {
					backgroundImage = currentBackgroundImage - foregroundMask;
					resultImage = foregroundImage | backgroundImage;
					medianBlur(resultImage,resultImage,3);
					imshow("Result!",resultImage);

					if(isRecording==true) {
						recordedMovie << resultImage;
					}
				}
			}
		}
	}
	return 0;
}